import os
import numpy as np
import csv
import myokit
import ionbench

class Staircase_Benchmarker(ionbench.benchmarker.Benchmarker):
    def __init__(self):
        self._log = myokit.DataLog.load_csv(os.path.join(ionbench.DATA_DIR, 'staircase', 'staircase-ramp.csv'))
        self._trueParams = np.copy(self.defaultParams)
        try:
            self.load_data(os.path.join(ionbench.DATA_DIR, 'staircase', 'data'+self._modelType+'.csv'))
        except FileNotFoundError:
            self.data = None
        self.sim = myokit.Simulation(self.model)
        self.sim.set_tolerance(1e-8,1e-8)
        self.add_protocol()
        super().__init__()
    
    def sample(self, n=1):
        """
        Sample parameters for the staircase problems.

        Parameters
        ----------
        n : int, optional
            Number of parameter vectors to sample. The default is 1.

        Returns
        -------
        params : list
            If n=1, then params is the vector of parameters. Otherwise, params is a list containing n parameter vectors.

        """
        params = [None]*n
        for i in range(n):
            params[i] = self.input_parameter_space(self.defaultParams*np.random.uniform(0.5,1.5,self.n_parameters()))
        if n==1:
            return params[0]
        else:
            return params
    
    def add_protocol(self):
        protocol = myokit.TimeSeriesProtocol(self._log.time(), self._log['voltage'])
        self.sim.set_protocol(protocol)
        self.tmax = self._log.time()[-1]
        self.sim.pre(500) #Prepace for 500ms
        
class HH_Benchmarker(Staircase_Benchmarker):
    """
    The Hodgkin-Huxley IKr Staircase benchmarker. 
    
    The benchmarker uses the Beattie et al 2017 IKr Hodgkin-Huxley model using the staircase protocol. 
    
    Its parameters are specified as scaling factors, so start at a vector of all ones, or sample from the benchmarker.sample() method. 
    """
    def __init__(self):
        print('Initialising Hodgkin-Huxley IKr benchmark')
        self._name = "staircase.hh"
        self.model = myokit.load_model(os.path.join(ionbench.DATA_DIR, 'staircase', 'beattie-2017-ikr-hh.mmt'))
        self._outputName = 'ikr.IKr'
        self._paramContainer = 'ikr'
        self._modelType = 'HH'
        self.defaultParams = np.array([2.26e-4, 0.0699, 3.45e-5, 0.05462, 0.0873, 8.91e-3, 5.15e-3, 0.03158, 0.1524])
        self._rateFunctions = [(lambda p,V:p[0]*np.exp(p[1]*V), 'positive'), (lambda p,V:p[2]*np.exp(-p[3]*V), 'negative'), (lambda p,V:p[4]*np.exp(p[5]*V), 'positive'), (lambda p,V:p[6]*np.exp(-p[7]*V), 'negative')] #Used for rate bounds
        self.standardLogTransform = [True, False]*4+[False]
        super().__init__()
        print('Benchmarker initialised')

class MM_Benchmarker(Staircase_Benchmarker):
    """
    The Markov IKr Staircase benchmarker. 
    
    The benchmarker uses the Fink et al 2008 IKr Markov model using the staircase protocol. 
    
    Its parameters are specified as scaling factors, so start at a vector of all ones, or sample from the benchmarker.sample() method. 
    """
    def __init__(self):
        print('Initialising Markov Model IKr benchmark')
        self._name = "staircase.mm"
        self.model = myokit.load_model(os.path.join(ionbench.DATA_DIR, 'staircase', 'fink-2008-ikr-mm.mmt'))
        self._outputName = 'IKr.i_Kr'
        self._paramContainer = 'iKr_Markov'
        self._modelType = 'MM'
        self.defaultParams = np.array([0.20618, 0.0112, 0.04209, 0.02202, 0.0365, 0.41811, 0.0223, 0.13279, 0.0603, 0.08094, 0.0002262, 0.0399, 0.04150, 0.0312])
        self._rateFunctions = [(lambda p,V:p[0]*np.exp(p[1]*V), 'positive'), (lambda p,V:p[2], 'independent'), (lambda p,V:p[3]*np.exp(p[4]*V), 'positive'), (lambda p,V:p[5]*np.exp(p[6]*V), 'positive'), (lambda p,V:p[7]*np.exp(-p[8]*V), 'negative'), (lambda p,V:p[9], 'independent'), (lambda p,V:p[10]*np.exp(-p[11]*V), 'negative'), (lambda p,V:p[12]*np.exp(-p[13]*V), 'negative')] #Used for rate bounds
        self.standardLogTransform = [True, False, True]*2+[False, True]*2+[True,False]*2
        super().__init__()
        print('Benchmarker initialised')

def generate_data(modelType):
    """
    Generate the data files for the staircase benchmarker problems.

    Parameters
    ----------
    modelType : string
        'HH' to generate the data for the Hodgkin-Huxley benchmark problem. 'MM' to generate the data for the Markov model benchmark problem.

    Returns
    -------
    None.

    """
    modelType = modelType.upper()
    if modelType == 'HH':
        bm = HH_Benchmarker()
    elif modelType == 'MM':
        bm = MM_Benchmarker()
    out = bm.simulate(bm._trueParams, np.arange(bm.tmax), continueOnError = False)
    out += np.random.normal(0, np.mean(np.abs(out))*0.05, len(out))
    with open(os.path.join(ionbench.DATA_DIR, 'staircase', 'data'+modelType+'.csv'), 'w', newline = '') as csvfile:
        writer = csv.writer(csvfile, delimiter = ',')
        writer.writerows(map(lambda x: [x], out))
    with open(os.path.join(ionbench.DATA_DIR, 'staircase', 'trueParams'+modelType+'.csv'), 'w', newline = '') as csvfile:
        writer = csv.writer(csvfile, delimiter = ',')
        writer.writerows(map(lambda x: [x], bm._trueParams))



import numpy as np
import myokit
import csv
import os
import ionbench
import matplotlib.pyplot as plt
import warnings
import pickle

class Tracker():
    """
    This class records the various performance metrics used to evaluate the optimisation algorithms. 
    
    It records the number of times the model is solved (stored as tracker.solveCount), not including the times that parameters were out of bounds.
    
    It records the RMSE (Root Mean Squared Error) cost each time a parameter vector is evaluated, using np.inf if parameters are out of bounds.
    
    It records the RMSRE (Root Mean Squared Relative Error) of the estimated parameters each time a parameter vector is evaluated. This is relative to the true parameters, ie the RMS of the vector of error between the estimated and true parameters, expressed as a percentage of the true parameters.
    
    It records the number of parameters that were correctly identified (within 5% of the true values) each time a parameter vector is evaluated.
    
    This class contains two methods: update(), and plot(). 
    
    update() is called everytime a parameter vector is simulated in the benchmarker (for example, in bm.cost) and updates the performance metric vectors.
    
    plot() is called during the benchmarkers .evaluate() method. It plots the performance metrics as functions of time (in the order in which parameter vectors were evaluated).
    """
    def __init__(self, trueParams):
        self.costs = []
        self.paramRMSRE = []
        self.paramIdentifiedCount = []
        self.solveCount = 0
        self.firstParams = []
        self.modelSolves = []
        self._trueParams = trueParams
    
    def update(self, estimatedParams, cost = np.inf, incrementSolveCounter = True):
        """
        This method updates the performance metric tracking vectors with new values. It should only need to be called by a benchmarker class.
        
        Parameters
        ----------
        
        estimatedParams : arraylike
            The vector of parameters that are being evaluated, after any transformations to return them to the original parameter space have been applied.
        cost : float, optional
            The RMSE cost of the parameter vectors that are being evaluated. The default is np.inf, to be used if parameters are out of bounds.
        incrementSolveCounter : bool, optional
            Should the solveCount be incremented, ie did the model need to be solved. This should only be False during benchmarker.evaluate() or if the parameters were out of bounds. The default is True.

        Returns
        -------
        None.

        """
        if len(self.firstParams)==0:
            self.firstParams = np.copy(estimatedParams)
        #Cast to numpy arrays
        trueParams = np.array(self._trueParams)
        estimatedParams = np.array(estimatedParams)
        #Update performance metrics
        self.paramRMSRE.append(np.sqrt(np.mean(((estimatedParams-trueParams)/trueParams)**2)))
        self.paramIdentifiedCount.append(np.sum(np.abs((estimatedParams-trueParams)/trueParams)<0.05))
        self.costs.append(cost)
        if incrementSolveCounter:
            self.solveCount += 1
        self.modelSolves.append(self.solveCount)
    
    def plot(self):
        """
        This method plots the performance metrics as functions of time (in the order in which parameter vectors were evaluated). It will produce three plots, the RMSE cost, parameter RMSRE, and the number of identified parameters over the optimisation. This method will be called when benchmarker.evaluate() is called, so long as benchmarker.plotter = True (the default).

        Returns
        -------
        None.

        """
        #Cost plot
        plt.figure()
        plt.scatter(range(len(self.costs)),self.costs, c="k", marker=".")
        plt.xlabel('Cost function calls')
        plt.ylabel('RMSE cost')
        plt.title('Data error')
        
        #Parameter RMSRE plot
        plt.figure()
        plt.scatter(range(len(self.paramRMSRE)),self.paramRMSRE, c="k", marker=".")
        plt.xlabel('Cost function calls')
        plt.ylabel('Parameter RMSRE')
        plt.title('Parameter error')
        
        #Number of identified parameters plot
        plt.figure()
        plt.scatter(range(len(self.paramIdentifiedCount)),self.paramIdentifiedCount, c="k", marker=".")
        plt.xlabel('Cost function calls')
        plt.ylabel('Number of parameters identified')
        plt.title('Number of parameters identified')
    
    def save(self, filename):
        """
        Saves the tracked variables. Useful to store results to plot later.

        Parameters
        ----------
        filename : string
            Filename for storing the tracked variables (solveCount, costs, modelSolves, paramRMSRE, paramIdentifiedCount). Variables will be pickled and stored in working directory.

        Returns
        -------
        None.

        """
        data = (self.solveCount, self.costs, self.modelSolves, self.paramRMSRE, self.paramIdentifiedCount)
        with open(filename, 'wb') as f:
            pickle.dump(data,f)
    
    def load(self, filename):
        """
        Loads the tracked variables. Useful to store results to plot later.

        Parameters
        ----------
        filename : string
            Filename to load the stored tracked variables (solveCount, costs, modelSolves, paramRMSRE, paramIdentifiedCount). Variables will be read for the working directory. Must be saved using the .save() method associated with a tracker.

        Returns
        -------
        None.

        """
        with open(filename, 'rb') as f:
            data = pickle.load(f)
        self.solveCount, self.costs, self.modelSolves, self.paramRMSRE, self.paramIdentifiedCount = data
    
    def report_convergence(self):
        """
        Reports the performance metrics at the point of convergence, defined as the first point where there is an unbroken chain in the number of correctly identified parameters.

        Returns
        -------
        None.

        """
        finalParamId = self.paramIdentifiedCount[-1]
        ifEqualFinalParamId = self.paramIdentifiedCount == finalParamId
        ind = [i for i, x in enumerate(ifEqualFinalParamId) if x] #Indexes where number of parameters identified is equal to the final count
        for i in ind:
            if all(ifEqualFinalParamId[i:]):
                #All future points remain with this many parameters identified, therefore it is considered converged
                print('Model solves until convergence:  '+str(self.modelSolves[i]))
                print('Cost at convergence:             {0:.6f}'.format(self.costs[i]))
                print('Parameter RMSRE at convergence:  {0:.6f}'.format(self.paramRMSRE[i]))
                break
    
class Benchmarker():
    """
    The Benchmarker class contains all the features needed to evaluate an optimisation algorithm. This class should not need to be called directly and is instead used as a parent class for the benchmarker problems. 
    
    The main methods to use from this class are n_parameters(), cost(), reset(), and evaluate().
    """
    def __init__(self):
        self._useScaleFactors = False
        self._bounded = False #Should the parameters be bounded
        self._logTransformParams = [False]*self.n_parameters() #Are any of the parameter log-transformed
        self.plotter = True #Should the performance metrics be plotted when evaluate() is called
        self.tracker = Tracker(self._trueParams) #Tracks the performance metrics
        
    def load_data(self, dataPath = '', paramPath = ''):
        """
        Loads output data to use in fitting.

        Parameters
        ----------
        dataPath : string, optional
            An absolute filepath to the .csv data file. The default is '', in which case no file will be loaded.
        paramPath : string, optional
            An absolute filepath to the .csv file containing the true parameters. The default is '', in which case no file will be loaded.

        Returns
        -------
        None.

        """
        if not dataPath == '':
            tmp=[]
            with open(dataPath, newline='') as csvfile:
                reader = csv.reader(csvfile)
                for row in reader:
                    tmp.append(float(row[0]))
            self.data = np.array(tmp)
        if not paramPath == '':
            tmp=[]
            with open(paramPath, newline='') as csvfile:
                reader = csv.reader(csvfile)
                for row in reader:
                    tmp.append(float(row[0]))
            self._trueParams = np.array(tmp)
    
    def add_bounds(self, bounds, parameterSpace = 'original'):
        """
        Add bounds to the parameters. The bounds will be checked whenever the model is about to be solved, if they are violated then the model will not be solved and an infinite cost will be reported. The model solve count will not be incremented if the bounds are violated but the parameter vector will still be tracked for the other metrics.
        
        The bounds are checked after reversing any log-transforms on parameters (ie on exp(inputted parameters)).

        Parameters
        ----------
        bounds : list
            A list of bounds. The list should contain two elements, the first is a list of lower bounds, the same length as a parameter vector and the second a similar list of upper bounds. Use -np.inf or np.inf to not include bounds on particular parameters.
        parameterSpace : string
            Specify the parameter space of the inputted bounds. The options are 'original', to specify the bounds in the original parameter space, using the parameters that will be evaluated in the model, or 'input' for the parameter space inputted by the cost function, meaning they will be scaled by the default parameters if benchmarker._useScaleFactors=True, and log transformed if any parameters are to be log transformed. This will only have a difference if parameters are log transformed or if  benchmarker._useScaleFactors=True. The default is 'original'.

        Returns
        -------
        None.

        """
        if parameterSpace.lower() == 'input':
            self.lb = self.original_parameter_space(bounds[0])
            self.ub = self.original_parameter_space(bounds[1])
        elif parameterSpace.lower() == 'original':
            self.lb = bounds[0]
            self.ub = bounds[1]
        self._bounded = True
    
    def log_transform(self, whichParams = []):
        """
        Fit some parameters in a log-transformed space. 
        
        Inputted log-transformed parameters will be set to exp(inputted parameters) before solving the model.

        Parameters
        ----------
        whichParams : list, optional
            Which parameters should be log-transformed, in the form of a list of bools, the same length as the number of parameters, where True is a parameter to be log-transformed. The default is [], in which case all parameters will be log-transformed.

        Returns
        -------
        None.

        """
        if whichParams == []: #Log-transform all parameters
            whichParams = [True]*self.n_parameters()
        self._logTransformParams = whichParams
    
    def input_parameter_space(self, parameters):
        """
        Maps parameters from the original parameter space to the input space. Incorporating any log transforms or scaling factors.

        Parameters
        ----------
        parameters : list
            Parameter vector in the original parameter space.

        Returns
        -------
        parameters : list
            Parameter vector mapped to input space.

        """
        parameters = np.copy(parameters)
        for i in range(self.n_parameters()):
            if self._useScaleFactors:
                parameters[i] = parameters[i]/self.defaultParams[i]
            if self._logTransformParams[i]:
                parameters[i] = np.log(parameters[i])
        
        return parameters
    
    def original_parameter_space(self, parameters):
        """
        Maps parameters from input space to the original parameter space. Removing any log transforms or scaling factors.

        Parameters
        ----------
        parameters : list
            Parameter vector in input space.

        Returns
        -------
        parameters : list
            Parameter vector mapped to the original parameter space.

        """
        parameters = np.copy(parameters)
        for i in range(self.n_parameters()):
            if self._logTransformParams[i]:
                parameters[i] = np.exp(parameters[i])
            if self._useScaleFactors:
                parameters[i] = parameters[i]*self.defaultParams[i]
        
        return parameters
    
    def in_bounds(self, parameters):
        """
        Checks if parameters are inside any bounds. If benchmarker._bounded = False, then it always returns True.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters (in original parameter space) to check against the bounds.

        Returns
        -------
        bool
            True if parameters are inside the bounds or no bounds are specified, False if the parameters are outside of the bounds.

        """
        if self._bounded:
            if any(parameters[i]<self.lb[i] or parameters[i]>self.ub[i] for i in range(self.n_parameters())):
                return False
        return True
    
    def n_parameters(self):
        """
        Returns the number of parameters in the model.
        """
        return len(self.defaultParams)
    
    def reset(self):
        """
        Resets the benchmarker. This clears the simulation object and restarts the performance tracker.

        Returns
        -------
        None.

        """
        self.sim.reset()
        self.tracker = Tracker(self._trueParams)
    
    def cost(self, parameters, incrementSolveCounter = True):
        """
        Find the RMSE cost between the model solved using the inputted parameters and the data.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.
        incrementSolveCounter : bool, optional
            If False, it disables the solve counter tracker. This never needs to be set to False by a user. This is only required by the evaluate() method. The default is True.

        Returns
        -------
        cost : float
            The RMSE cost of the parameters.
        
        """
        testOutput = np.array(self.simulate(parameters, np.arange(0, self.tmax), incrementSolveCounter = incrementSolveCounter))
        cost = np.sqrt(np.mean((testOutput-self.data)**2))
        return cost
    
    def signed_error(self, parameters):
        """
        Similar to the cost method, but instead returns the vector of residuals/errors in the model output.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.

        Returns
        -------
        signed_error : numpy array
            The vector of model errors/residuals.

        """
        #Calculate cost for a given set of parameters
        testOutput = np.array(self.simulate(parameters, np.arange(0, self.tmax)))
        return (testOutput-self.data)
    
    def squared_error(self, parameters):
        """
        Similar to the cost method, but instead returns the vector of squared residuals/errors in the model output.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.

        Returns
        -------
        signed_error : numpy array
            The vector of model squared errors/residuals.
        
        """
        #Calculate cost for a given set of parameters
        testOutput = np.array(self.simulate(parameters, np.arange(0, self.tmax)))
        return (testOutput-self.data)**2
    
    def set_params(self, parameters):
        """
        Set the parameters in the simulation object. Inputted parameters should be in the original parameter space.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.

        Returns
        -------
        None.

        """
        # Update the parameters
        for i in range(self.n_parameters()):
            self.sim.set_constant(self._paramContainer+'.p'+str(i+1), parameters[i])
    
    def solve_model(self, times, continueOnError = True):
        """
        Solve the model at the inputted times and return the current trace.

        Parameters
        ----------
        times : list or numpy array
            Vector of timepoints to record model output. Typically in ms.
        continueOnError : bool, optional
            If continueOnError is True, any errors that occur during solving the model will be ignored and an infinite output will be given. The default is True.

        Returns
        -------
        modelOutput : list
            A vector of model outputs (current trace).
        
        """
        if continueOnError:
            try:
                log = self.sim.run(self.tmax+1, log_times = times)
                return np.array(log[self._outputName])
            except:
                warnings.warn("Failed to solve model. Will report infinite output in the hope of continuing the run.")
                return np.array([np.inf]*len(times))
        else:
            log = self.sim.run(self.tmax+1, log_times = times)
            return np.array(log[self._outputName])
    
    def simulate(self, parameters, times, continueOnError = True, incrementSolveCounter = True):
        """
        Simulate the model for the inputted parameters and return the model output at the specified times.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.
        times : list or numpy array
            Vector of timepoints to record model output. Typically in ms.
        continueOnError : bool, optional
            If continueOnError is True, any errors that occur during solving the model will be ignored and an infinite output will be given. The default is True.
        incrementSolveCounter : bool, optional
            If False, it disables the solve counter tracker. This never needs to be set to False by a user. This is only required by the evaluate() method. The default is True.

        Returns
        -------
        modelOutput : list
            A vector of model outputs (current trace).

        """
        #Return the parameters to the original parameter space
        parameters = self.original_parameter_space(parameters) #Creates a copy of the parameter vector
        
        # Reset the simulation
        self.sim.reset()
        
        # Abort solving if the parameters are out of bounds
        if not self.in_bounds(parameters):
            self.tracker.update(parameters, incrementSolveCounter = False)
            return [np.inf]*len(times)
        
        # Set the parameters in the simulation object
        self.set_params(parameters)
        
        # Run the simulation and track the performance
        out = self.solve_model(times, continueOnError = continueOnError)
        self.tracker.update(parameters, cost = np.sqrt(np.mean((out-self.data)**2)), incrementSolveCounter = incrementSolveCounter)
        return out
    
    def evaluate(self, parameters):
        """
        Evaluates a final set of parameters. 
        
        This method reports the performance metrics for this parameter vector (calling evaluate() does NOT increase the number of cost function evaluations). If benchmarker.plotter = True, then it also plots the performance metrics over the course of the optimisation.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model, evaluating the performance of the final parameter vector.

        Returns
        -------
        None.

        """
        print('')
        print('=========================================')
        print('===    Evaluating Final Parameters    ===')
        print('=========================================')
        print('')
        print('Number of cost evaluations:      '+str(self.tracker.solveCount))
        cost =  self.cost(parameters, incrementSolveCounter = False)
        print('Final cost:                      {0:.6f}'.format(cost))
        print('Parameter RMSRE:                 {0:.6f}'.format(self.tracker.paramRMSRE[-1]))
        print('Number of identified parameters: '+str(self.tracker.paramIdentifiedCount[-1]))
        print('Total number of parameters:      '+str(self.n_parameters()))
        self.tracker.report_convergence()
        print('')
        if self.plotter:
            self.tracker.plot()
            self.sim.reset()
            self.set_params(self.tracker.firstParams)
            firstOut = self.solve_model(np.arange(self.tmax), continueOnError = True)
            self.sim.reset()
            self.set_params(self.original_parameter_space(parameters))
            lastOut = self.solve_model(np.arange(self.tmax), continueOnError = True)
            plt.figure()
            if "moreno" in self._name:
                plt.plot(self.data)
                plt.plot(firstOut)
                plt.plot(lastOut)
                plt.ylabel('Summary Statistics')
            else:
                plt.plot(np.arange(self.tmax),self.data)
                plt.plot(np.arange(self.tmax),firstOut)
                plt.plot(np.arange(self.tmax),lastOut)
                plt.ylabel('Current')
                plt.xlabel('Time (ms)')
            plt.legend(['Data','First Parameters','Final Parameters'])
            plt.title('Improvement after fitting: '+self._name)
            plt.show()



import ionbench.problems.staircase
import scipy.optimize

def run(bm, x0 = [], diff_step = 1e-3, maxfev = 20000):
    """
    Runs Trust Region Reflective optimiser from Scipy.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is [], in which case a randomly sampled parameter vector is retrieved from bm.sample().
    diff_step : float, optional
        Step size for finite difference calculation. The default is 1e-3.
    maxfev : int, optional
        Maximum number of cost function evaluations. The default is 20000.

    Returns
    -------
    xbest : list
        The best parameters identified by Trust Region Reflective.

    """
    if len(x0)==0:
        x0 = bm.sample()
    
    if bm._bounded:
        bounds = (bm.lb,bm.ub)
        out = scipy.optimize.least_squares(bm.signed_error, x0, method='trf', diff_step=diff_step, verbose=2, max_nfev = maxfev, bounds = bounds)
    else:
        out = scipy.optimize.least_squares(bm.signed_error, x0, method='trf', diff_step=diff_step, verbose=2, max_nfev = maxfev)
    
    bm.evaluate(out.x)
    return out.x

if __name__ == '__main__':
    bm = ionbench.problems.staircase.HH_Benchmarker()
    run(bm)

def get_approach():
    """
    No approach specified. Will use an empty approach

    Returns
    -------
    app : approach
        Empty approach

    """
    app = ionbench.approach.Empty(name = 'trustRegionReflective_scipy')
    return app



import numpy as np
import ionbench
import itertools
import copy
import matplotlib.pyplot as plt
import pickle

def secondDeriv(bm, i, j, step = 1e-4, buffer = 1e-4):
    """
    Calculate the second order partial derivative of the log likelihood (specifically the negative of the sum of squared errors) in parameters i and j. 

    Parameters
    ----------
    bm : benchmarker
        A benchmarker to define the log likelihood.
    i : int
        Index of the first parameter for the derivative.
    j : int
        Index of the second parameter for the derivative.
    step : float, optional
        The finite difference step size. If the step size is deemed too small, it will increase by step until the points used by finite difference are at least buffer away from the maximum likelihood. The default is 1e-4.
    buffer : float, optional
        While the evaluated negative sum of squared errors is within buffer of that at the maximum likelihood, the step size will be increased. The value of buffer should be at least 30 times larger than the noise caused by the ODE solver. The default is 1e-4.

    Returns
    -------
    deriv : float
        The second derivative of the negative sum of squared errors.

    """
    if i==j:
        def f(x):
            p = bm.input_parameter_space(bm._trueParams)
            p[i] = x
            return sse(bm, p)
        #fxx = f(x+h,y)-2f(x,y)+f(x-h,y) / h^2
        #Find h st f(x+h,y) < f(x,y)-buffer and f(x-h,y) < f(x,y)-buffer
        h = step
        x = bm.input_parameter_space(bm._trueParams)[i]
        centre = f(x)
        up = f(x+x*h)
        down = f(x-x*h)
        while up >= centre-buffer or down >= centre-buffer:
            h += step
            up = f(x+x*h)
            down = f(x-x*h)
        return (up-2*centre+down)/((x*h)**2)
    else:
        def f(x,y):
            p = bm.input_parameter_space(bm._trueParams)
            p[i] = x
            p[j] = y
            return sse(bm, p)
        #fxy = f(x+h,y+k)-f(x+h,y-k)-f(x-h,y+k)+f(x-h,y-k) / 4hk
        #Find h st all of f(x+h,y+h), f(x-h,y+h), f(x+h,y-h), f(x-h,y-h) < f(x,y)-buffer
        h = step
        x = bm.input_parameter_space(bm._trueParams)[i]
        y = bm.input_parameter_space(bm._trueParams)[j]
        centre = f(x,y)
        up_up = f(x+x*h,y+y*h)
        up_down = f(x+x*h,y-y*h)
        down_up = f(x-x*h,y+y*h)
        down_down = f(x-x*h,y-y*h)
        while up_up >= centre-buffer or up_down >= centre-buffer or down_up >= centre-buffer or down_down >= centre-buffer:
            h += step
            up_up = f(x+x*h,y+y*h)
            up_down = f(x+x*h,y-y*h)
            down_up = f(x-x*h,y+y*h)
            down_down = f(x-x*h,y-y*h)
        return (up_up-up_down-down_up+down_down)/(4*x*h*y*h)

def sse(bm, params):
    """
    Negative sum of squared errors.

    Parameters
    ----------
    bm : benchmarker
        Defines the cost function, solver and data.
    params : list
        List of parameters at which to find the negative sum of squared errors.

    Returns
    -------
    sse : float
        Negative sum of squared errors.

    """
    cost = bm.cost(params)
    return -cost**2*len(bm.data) #Sum of squared errors

def run(bm, sigma = 1, preoptimise = True, ftol = 3e-6, step = 1e-4, buffer = 1e-4):
    """
    Calculate the Fisher's Information Matrix for a benchmarker problem. Also plots the eigenspectrum of the FIM.

    Parameters
    ----------
    bm : benchmarker
        A benchmarker of which to find the FIM matrix.
    sigma : float, optional
        The variance of the noise in the benchmarker's data. The default is 1.
    preoptimise : bool, optional
        Should the true parameters be preoptimised to ensure they are the MLE, useful due to biases from noise. If so, then Scipy's Nelder Mead will be run starting from the true parameters. The default is True.
    ftol : float, optional
        The cost function tolerance to use for the Nelder Mead preoptimisation. Should be slightly larger than the noise in the solver. The default is 3e-6.
    step : float, optional
        Used in finite differences. The finite difference step size. If the step size is deemed too small, it will increase by step until the points used by finite difference are at least buffer away from the maximum likelihood. The default is 1e-4.
    buffer : float, optional
        Used in finite differences. While the evaluated negative sum of squared errors is within buffer of that at the maximum likelihood, the step size will be increased. The value of buffer should be at least 30 times larger than the noise caused by the ODE solver. The default is 1e-4.
    
    Returns
    -------
    mat : numpy matrix
        The Fisher's Information Matrix.

    """
    mat = np.zeros((bm.n_parameters(), bm.n_parameters()))
    
    bm._useScaleFactors = True
    #Search from the true parameters to find the actual MLE
    if preoptimise:
        out = ionbench.optimisers.scipy_optimisers.nelderMead_scipy.run(bm, x0=bm.input_parameter_space(bm._trueParams),ftol=ftol)
        bm._trueParams = bm.original_parameter_space(out)
        print('Identified MLE at:')
        print(bm._trueParams)
        print('Storing in this benchmarker object as the true parameters')
    for i, j in itertools.combinations(range(bm.n_parameters()),2):
        d = secondDeriv(bm, i, j, step = step, buffer = buffer)*1/(2*sigma**2)
        mat[i,j] = d
        mat[j,i] = d
    for i in range(bm.n_parameters()):
        mat[i,i] = secondDeriv(bm, i, i, step = step, buffer = buffer)*1/(2*sigma**2)
    eigs = np.linalg.eigvals(mat)
    plt.eventplot(-eigs, orientation='vertical')
    plt.yscale('log')
    plt.xticks(ticks=[])
    plt.title('FIM Eigenspectrum: '+bm._name)
    data = (eigs, mat)
    with open(bm._name+'_fim.pickle', 'wb') as f:
        pickle.dump(data,f)
    return mat

def explore_solver_noise(bm):
    """
    Identify the scale of noise in the solver cost function.

    Parameters
    ----------
    bm : benchmarker
        Defines the cost function and model solver.

    Returns
    -------
    None.

    """
    bm._useScaleFactors = True
    p = bm.input_parameter_space(bm._trueParams)
    params = [copy.copy(p) for i in range(100)]
    for i in range(100):
        params[i][0] = 0.9999999**i
    #print(params)
    sses = [sse(bm,a) for a in params]
    plt.figure()
    plt.plot(sses)
    plt.show()
    print('STD: '+str(np.std(sses)))
    print('Range: '+str(max(sses)-min(sses)))



import ionbench
import numpy as np

class Approach():
    """
    Approaches provide a way to store some benchmarker settings and apply them to a range of benchmarkers. These settings include log-transforms, parameter bounds, and the use of scale factors. A single approach can be generated, and then applied to a range of benchmarkers. Once a Approach object is instantiated, its transforms and bounds can be applied to a benchmarker, bm, by calling approach.apply(bm)
    """
    def __init__(self, name = "", logTransform = 'None', bounds = 'None', scaleFactors = 'off', customLogTransform = None, customBounds = None):
        """
        Initialise an Approach object. Once an Approach is built, its setting can be applied to any benchmarker, bm, by calling approach.apply(bm)

        Parameters
        ----------
        name : string, optional
            A name for the approach. Useful for logging. The default is "".
        logTransform : string, optional
            Setting for log transforms. Options are 'None' (default, none of the parameters will be log transformed), 'Standard' (a problem specific set of parameters will be log transformed), 'Full' (all parameters will be log transformed), and 'Custom' (Log transformed parameters specified by the customLogTransform input will be log transformed).
        bounds : string, optional
            Setting for parameter bounds. Options are 'None' (default, none of the parameters will be bounds), 'Positive' (all parameters will have a lower bound of 0 and no upper bound), 'Sampler' (problem specific bounds using the range in the benchmarkers sampler will be used), and 'Custom' (Bounds on parameters specified by the customBounds input will be used).
        scaleFactors : string, optional
            Setting for scale factors. Options are 'off' (default, scale factors won't be used), and 'on' (scale factors will be used, with the default parameters representing 1).
        customLogTransform : list, optional
            A list of parameters to log transform. The list should be the same length as the number of parameters in the benchmarker. This makes this option problem specific and will be unlikely to be useable across different benchmarkers. Only required if logTransform = 'Custom'. 
        customBounds : list, optional
            A list containing 2 elements, a list of lower bounds and a list of upper bounds. Each sub-list should be the same length as the number of parameters in the benchmarker. This makes this option problem specific and will be unlikely to be useable across different benchmarkers. Only required if bounds = 'Custom'. 

        Returns
        -------
        None.

        """
        self.dict = {'log transform':logTransform, 'bounds':bounds, 'scale factors':scaleFactors}
        self.customLogTransform = customLogTransform
        self.customBounds = customBounds
        self._name = None
    
    def apply(self, bm):
        """
        Applies the settings in this approach to the inputted benchmarker. This will override any previous settings assigned to the benchmarker.

        Parameters
        ----------
        bm : benchmarker
            A benchmarker problem to apply this approach.

        Returns
        -------
        None.

        """
        self.apply_log_transforms(self.dict['log transform'], bm)
        self.apply_bounds(self.dict['bounds'], bm)
        self.apply_scale_factors(self.dict['scale factors'], bm)
    
    def apply_log_transforms(self, setting, bm):
        """
        Apply the log transform setting to a benchmarker.

        Parameters
        ----------
        setting : string
            Setting for log transforms. Options are 'None' (default, none of the parameters will be log transformed), 'Standard' (a problem specific set of parameters will be log transformed), 'Full' (all parameters will be log transformed), and 'Custom' (Log transformed parameters specified by the customLogTransform input will be log transformed, only usable if customLogTransform was set during the creation of this approach, or approach.customLogTransform is set).
        bm : benchmarker
            A benchmarker problem to apply this approach.

        Returns
        -------
        None.

        """
        if setting.lower() == 'none':
            #Disable log transforms
            bm.log_transform([False]*bm.n_parameters())
        elif setting.lower() == 'standard':
            #Load standard log transforms from benchmarker
            bm.log_transform(bm.standardLogTransform)
        elif setting.lower() == 'full':
            #Log transform all parameters
            bm.log_transform()
        elif setting.lower() == 'custom':
            bm.log_transform(self.customLogTransform)
        else:
            print("'"+setting+"' is not a valid option for log transforms. Please use either 'None', 'Standard', 'Full', or 'Custom'.")
    
    def apply_bounds(self, setting, bm):
        """
        Apply the bound setting to a benchmarker.

        Parameters
        ----------
        setting : string
            Setting for parameter bounds. Options are 'None' (default, none of the parameters will be bounds), 'Positive' (all parameters will have a lower bound of 0 and no upper bound), 'Sampler' (problem specific bounds using the range in the benchmarkers sampler will be used), and 'Custom' (Bounds on parameters specified by the customBounds input will be used). The default is 'None'.
        bm : benchmarker
            A benchmarker problem to apply this approach.

        Returns
        -------
        None.

        """
        if setting.lower() == 'none':
            bm.add_bounds([[-np.inf]*bm.n_parameters(),[np.inf]*bm.n_parameters()])
            bm._bounded = False
        elif setting.lower() == 'positive':
            bm.add_bounds([[0]*bm.n_parameters(),[np.inf]*bm.n_parameters()])
        elif setting.lower() == 'sampler':
            if 'staircase' in bm._name:
                bm.add_bounds([bm.defaultParams*0.5,bm.defaultParams*1.5])
            elif 'loewe2016' in bm._name:
                lb = []
                ub = []
                for i in range(bm.n_parameters()):
                    if bm.additiveParams[i]:
                        lb.append(bm.defaultParams[i] - 60*bm.paramSpaceWidth)
                        ub.append(bm.defaultParams[i] + 60*bm.paramSpaceWidth)
                    else:
                        lb.append(bm.defaultParams[i]*10**-1)
                        ub.append(bm.defaultParams[i]*10**1)
                bm.add_bounds([lb,ub])
            elif 'moreno2016' in bm._name:
                bm.add_bounds([bm.defaultParams*(1-bm.paramSpaceWidth), bm.defaultParams*(1+bm.paramSpaceWidth)])
            else:
                print("Could not identify the benchmaker using benchmarker._name when trying to add bounds. If a new benchmarker is being used, then this code needs updating. For the time being, you can use 'custom' bounds.")
        elif setting.lower() == 'custom':
            bm.add_bounds(self.customBounds)
        else:
            print("'"+setting+"' is not a valid option for bounds. Please use either 'None', 'Positive', 'Sampler', or 'Custom'.")
    
    def apply_scale_factors(self, setting, bm):
        """
        Apply the scale factor setting to a benchmarker.

        Parameters
        ----------
        setting : string
            Setting for scale factors. Options are 'off' (default, scale factors won't be used), and 'on' (scale factors will be used, with the default parameters representing 1).
        bm : benchmarker
            A benchmarker problem to apply this approach.

        Returns
        -------
        None.

        """
        if setting.lower() == 'on':
            #Set use scale factors
            bm._useScaleFactors = True
        elif setting.lower() == 'off':
            #Disable scale factors
            bm._useScaleFactors = False
        else:
            print("'"+setting+"' is not a valid option for scale factors. Please use either 'on' or 'off'.")

class Clerx2019(Approach):
    """
    The approach from Clerx et al 2019. Uses standard log transforms, bounds defined by the sampler in place of rates bounds, and no scale factors.
    """
    def __init__(self):
        logTransform = 'Standard'
        bounds = 'Sampler'
        scaleFactors = 'off'
        name = 'Clerx2019'
        super().__init__(name = name, logTransform = logTransform, bounds = bounds, scaleFactors = scaleFactors)

class Loewe2016(Approach):
    """
    The approach from Loewe et al 2016. Uses no log transforms, bounds defined by the sampler, and no scale factors.
    """
    def __init__(self):
        logTransform = 'None'
        bounds = 'Sampler'
        scaleFactors = 'off'
        name = 'Loewe2012'
        super().__init__(name = name, logTransform = logTransform, bounds = bounds, scaleFactors = scaleFactors)

class Bot2012(Approach):
    """
    The approach from Bot et al 2012. Uses no log transforms, bounds defined by the sampler, and no scale factors.
    """
    def __init__(self):
        logTransform = 'None'
        bounds = 'Sampler'
        scaleFactors = 'off'
        name = 'Bot2012'
        super().__init__(name = name, logTransform = logTransform, bounds = bounds, scaleFactors = scaleFactors)

class Kohjitani2022(Approach):
    """
    The approach from Kohjitani et al 2022. Uses scaling factors only.
    """
    def __init__(self):
        logTransform = 'None'
        bounds = 'None'
        scaleFactors = 'on'
        name = 'Kohjitani2022'
        super().__init__(name = name, logTransform = logTransform, bounds = bounds, scaleFactors = scaleFactors)

class Chen2012(Approach):
    """
    The approach from Chen et al 2012. Uses sampler bounds only.
    """
    def __init__(self):
        logTransform = 'None'
        bounds = 'Sampler'
        scaleFactors = 'off'
        name = 'Chen2012'
        super().__init__(name = name, logTransform = logTransform, bounds = bounds, scaleFactors = scaleFactors)

class Empty(Approach):
    """
    An approach with default settings.
    """
    def __init__(self, name):
        logTransform = 'None'
        bounds = 'None'
        scaleFactors = 'off'
        super().__init__(name = name, logTransform = logTransform, bounds = bounds, scaleFactors = scaleFactors)



import numpy as np
import ionbench
import itertools
#Notes: The algorithm defines parameters between 0 and 1, this is mapped to 0 to 2 when the cost function is called

def run(bm, groups = [], n=20, c1=1.4, c2=1.4, qmax=5, lmax=200, gmin=0.05, w=0.6, debug=False):
    """
    Runs the perturbed particle swarm optimisation algorithm from Chen et al 2012. If the benchmarker is bounded, the solver will search in the interval [lb,ub], otherwise the solver will search in the interval [0,2*default]

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    groups : list, optional
        Groupings of parameters to use in the algorithm. Default is [], in which case every parameter will be in its own group.
    n : int, optional
        Number of particles. The default is 20.
    c1 : float, optional
        Scale of the acceleration towards a particles best positions. The default is 1.4.
    c2 : float, optional
        Scale of the acceleration towards the best point seen across all particles. The default is 1.4.
    qmax : int, optional
        Maximum number of iterations without improvement before perturbations. The default is 5.
    lmax : int, optional
        Maximum number of iterations. The default is 200.
    gmin : float, optional
        Target cost. Once the cost is reduced below gmin, the optimisation terminates. The default is 0.05.
    w : float, optional
        Initial value for the inertia of the particles. The default is 0.6.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    
    class particle:
        def __init__(self, n_param):
            self.velocity = 0.1*np.random.rand(n_param)
            self.position = np.random.rand(n_param)
            self.bestCost = np.inf #Best cost of this particle
            self.bestPosition = np.copy(self.position) #Position of best cost for this particle
            self.currentCost = None
        
        def set_cost(self, cost):
            self.currentCost = cost
            if cost < self.bestCost:
                self.bestCost = cost
                self.bestPosition = np.copy(self.position)
    
    def cost_func(x):
        return bm.cost(transform(x))
    
    def transform(x):
        if bm._bounded:
            xTrans = bm.lb + x*(bm.ub-bm.lb) #Map x to [lb,ub]
        else:
            xTrans = x*2 #Map x from [0,1] to [0,2]
            if not bm._useScaleFactors:
                xTrans = xTrans*bm.defaultParams #Map to [0,2*default]
        return xTrans
    
    if len(groups) == 0:
        groups = [[i] for i in range(bm.n_parameters())]
    
    q = 0 #Number of generations without improvement
    #Generate patterns
    patterns = [] #All combinations of groups
    for i in range(len(groups)):
        for j in itertools.combinations(range(len(groups)),i+1):
            newPattern = []
            for k in j:
                newPattern += groups[k]
            patterns += [newPattern]
    N = len(patterns) #Number of patterns
    
    particleList = []
    for i in range(n):
        particleList.append(particle(bm.n_parameters()))
    
    Gcost = [np.inf]*lmax #Best cost ever
    Gpos = [None]*lmax #Position of best cost ever
    alpha = [None]*lmax
    for L in range(lmax):
        if L > 0:
            Gcost[L] = Gcost[L-1]
            Gpos[L] = Gpos[L-1]
        
        if debug:
            print('-------------')
            print("Begginning population: "+str(L))
            print("Best cost so far: "+str(Gcost[L]))
            print("Found at position: "+str(Gpos[L]))
        
        foundImprovement = False
        for p in particleList:
            cost = cost_func(p.position)
            p.set_cost(cost)
            if cost < Gcost[L]:
                Gcost[L] = cost
                Gpos[L] = np.copy(p.position)
                foundImprovement = True
        
        if foundImprovement:
            q = 0
            if debug:
                print("Found improvement")
                print("Best cost is now: "+str(Gcost[L]))
                print("Found at position: "+str(Gpos[L]))
        else:
            q += 1
            if debug:
                print("Didn't find an improvement")
                print("Current value of q: "+str(q))
        
        if Gcost[L] < gmin:
            print("Cost successfully minimised")
            print("Final cost of:")
            print(Gcost[L])
            print("found at:")
            print(Gpos[L])
            break
        
        if q>5*qmax:
            #Abort
            print("Too many iterations without improvement")
            print("Final cost of:")
            print(Gcost[L])
            print("found at:")
            print(Gpos[L])
            break
        
        if q>=qmax:
            if debug:
                print("q exceeds qmax so perturbing")
                print("q: "+str(q))
                print("qmax: "+str(qmax))
            #steps 5.1 and 5.2
            newParticleList = []
            bestNewCost = np.inf
            bestNewPosition = None
            for i in range(N):
                newParticle = particle(bm.n_parameters())
                newParticle.position = np.copy(Gpos[L])
                for j in patterns[i]:
                    newParticle.position[j] *= 1+(np.random.rand()-0.5)/40
                    if newParticle.position[j] > 1:
                        newParticle.position[j] = 1
                cost = cost_func(newParticle.position)
                newParticle.set_cost(cost)
                if cost<bestNewCost:
                    bestNewCost = cost
                    bestNewPosition = np.copy(newParticle.position)
                newParticleList.append(newParticle)
            if debug:
                print("Perturbed particles")
                print("Best new cost is: "+str(bestNewCost))
                print("Found at: "+str(bestNewPosition))
            if bestNewCost <= Gcost[L]:
                if debug:
                    print("Cost improved by perturbing")
                    print("Original best cost: "+str(Gcost[L]))
                    print("New best cost: "+str(bestNewCost))
                q = 0
                Gcost[L] = bestNewCost
                Gpos[L] = np.copy(bestNewPosition)
                worstCost = -np.inf
                worstPosition = None
                for p in particleList:
                    if p.currentCost > worstCost:
                        worstCost = p.currentCost
                        worstPosition = p.position
                #Found worst cost and position, now to find that particle
                if debug:
                    print("Adjusting worst particle")
                for p in particleList:
                    if p.currentCost == worstCost and all(p.position == worstPosition):
                        if debug:
                            print("Worst particle found")
                            print("Worst particle cost: "+str(worstCost))
                            print("Worst particle position: "+str(worstPosition))
                        p.bestCost = bestNewCost
                        p.bestPosition = np.copy(bestNewPosition)
                        if debug:
                            print("New best cost for worst particle: "+str(p.bestCost))
                            print("New best position for worst particle: "+str(p.bestPosition))
            else:
                if debug:
                    print("Cost wasn't improved by perturbing")
        
        #Step 6
        alpha[L] = 1/n*np.sum(np.abs([p.currentCost-Gcost[L] for p in particleList]))
        if L>=2:
            #Adapt inertia weight w from (17)
            w = np.exp(-alpha[L-1]/alpha[L-2])
        if debug:
            print("Updating inertia")
            print("w: "+str(w))
            print("alpha: "+str(alpha[L]))
        
        #Renew velocities according to (16)
        for p in particleList:
            localAcc = c1*np.random.rand()*(p.bestPosition-p.position)
            globalAcc = c2*np.random.rand()*(Gpos[L]-p.position)
            p.velocity = w*p.velocity + localAcc + globalAcc
        if debug:
            print("Velocities renewed")
        #Move positions according to (15) while maintaining [0,1] bounds
        for p in particleList:
            p.position += p.velocity
            for i in range(bm.n_parameters()):
                if p.position[i] < 0:
                    p.position[i] = 0
                elif p.position[i] > 1:
                    p.position[i] = 1
        
        if debug:
            print("Positions renewed")
            print("Finished population: "+str(L))
            print("Best cost so far: "+str(Gcost[L]))
            print("Found at position: "+str(Gpos[L]))

    bm.evaluate(transform(Gpos[L]))
    return transform(Gpos[L])

if __name__ == '__main__':
    groups = [[0,2,4,6],[1,3,5,7],[8]]
    bm = ionbench.problems.staircase.HH_Benchmarker()
    run(bm,groups,debug=True)

def get_approach():
    """
    The approach in Chen et al 2012 uses bounds. The method also uses scale factors in such a way that the range of parameters is bounded between 0 and 1, however, this is done automatically in the optimisers for bounded benchmarkers.

    Returns
    -------
    app : approach
        The approach used in Chen et al 2012.

    """
    app = ionbench.approach.Chen2012()
    return app



import numpy as np
import ionbench
from functools import lru_cache
import copy

from pymoo.core.individual import Individual
from pymoo.core.problem import Problem
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PolynomialMutation
from pymoo.core.population import Population

def run(bm, nGens = 50, eta_cross = 10, eta_mut = 20, popSize = 50, debug = False):
    """
    Runs the genetic algorithm from Bot et al 2012.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    nGens : int, optional
        The number of generations to run the optimisation algorithm for. The default is 50.
    eta_cross : float, optional
        Crossover parameter. The default is 10.
    eta_mut : float, optional
        Mutation parameter. The default is 20.
    popSize : int, optional
        The size of the population in each generation. The default is 50.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    class individual():
        def __init__(self):
            self.x = bm.sample()
            self.cost = None
        def find_cost(self):
            self.cost = cost_func(tuple(self.x))
    
    @lru_cache(maxsize=None)
    def cost_func(x):
        return bm.cost(x)
    
    pop = [None]*popSize
    for i in range(popSize):
        pop[i] = individual()
        pop[i].find_cost()
    
    for gen in range(nGens):
        minCost = np.inf
        for i in range(popSize):
            if pop[i].cost < minCost:
                minCost = pop[i].cost
                elite = copy.deepcopy(pop[i])
        if debug:
            print("------------")
            print("Gen "+str(gen))
            print("Best cost: "+str(minCost))
        #Tournement selection
        newPop = []
        for j in range(2):
            perm = np.random.permutation(popSize)
            for i in range(popSize//2):
                if pop[perm[2*i]].cost > pop[perm[2*i+1]].cost:
                    newPop.append(copy.deepcopy(pop[perm[2*i]]))
                else:
                    newPop.append(copy.deepcopy(pop[perm[2*i+1]]))
        pop = newPop #Population of parents
        #Crossover SBX
        newPop = []
        problem = Problem(n_var=bm.n_parameters(), xl=0.0, xu=2.0)
        for i in range(popSize//2):
            a, b = Individual(X=np.array(pop[2*i].x)), Individual(X=np.array(pop[2*i+1].x))
    
            parents = [[a, b]]
            off = SBX(prob=0.9, eta=eta_cross).do(problem, parents) #What is prob vs prob_var
            Xp = off.get("X")
            newPop.append(individual())
            newPop[-1].x = Xp[0]
            newPop.append(individual())
            newPop[-1].x = Xp[1]
        pop = newPop
        #Mutation
        mutation = PolynomialMutation(prob=0.1*bm.n_parameters(), eta=eta_mut)
        for i in range(popSize):
            ind = Population.new(X=[pop[i].x])
            off = mutation(problem, ind)
            pop[i].x = off.get("X")[0]
        if debug:
            print("Finishing gen "+str(gen))
        #Find costs
        for i in range(popSize):
            pop[i].find_cost()
        #Elitism
        maxCost = -np.inf
        for i in range(popSize):
            if pop[i].cost > maxCost:
                maxCost = pop[i].cost
                maxIndex = i
        pop[maxIndex] = copy.deepcopy(elite)
    
    minCost = np.inf
    for i in range(popSize):
        if pop[i].cost < minCost:
            minCost = pop[i].cost
            elite = pop[i]
    bm.evaluate(elite.x)
    return elite.x

if __name__ == '__main__':
    bm = ionbench.problems.staircase.HH_Benchmarker()
    run(bm, debug=True)

def get_approach():
    """
    The approach in Bot et al uses a bounded search space for fitting conductances. The bounds are specified as a window around the true values, such as +-90%, +-10%, or -90% to +200% (although these may just be sampling starting points). This is closest to matching the 'sampler' bounds setting. There is no mention of scaling factor or log transforms.

    Returns
    -------
    app : approach
        The approach used in Bot et al 2012.

    """
    app = ionbench.approach.Bot2012()
    return app



from ionbench.optimisers.external_optimisers import GA_Bot2012
from ionbench.optimisers.external_optimisers import GA_Smirnov2020
from ionbench.optimisers.external_optimisers import patternSearch_Kohjitani2022
from ionbench.optimisers.external_optimisers import ppso_Chen2012
from ionbench.optimisers.external_optimisers import hybridPSOTRR_Loewe2016



import numpy as np
import scipy
import ionbench

def run(bm, n=96, K=5, Lmax=250, phi1=2.05, phi2=2.05, debug=False):
    """
    Runs the hybrid particle swarm optimisation - trust region reflective algorithm from Loewe et al 2016. If the benchmarker is bounded, the solver will search in the interval [lb,ub], otherwise the solver will search in the interval [0,2*default].
    
    Notes on using this optimiser:
        It is unclear the specifics of the optimisation used in Loewe et al 2016. It does not describe the method for sampling the initial positions or velocities of the particles (we use uniform distribution between bounds for position and expect Loewe et al 2016 did similarly, and zeros for the initial velocities, waiting for them to be set by TRR). The initial velocities may be different to those used in the Loewe et al 2016, but luckily the algorithm quickly overwrites the velocities after the first iteration based on the movement under TRR so it is only a difference between randomly sampled points and running TRR (our method) or randomly sampling points, moving them in a random direction and applying the bounds then applying TRR (a method with a non-zero random velocity).
        The TRR method in Loewe et al 2016 using Matlab's lsqnonlin implementation. As a Python based substitute we have used Scipy. Unfortunately, the Scipy implementation does not allow a maximum number of iterations to be specified. We have specified the maximum number of cost function evaluations (which appears to not include calls to calculate the gradient), which is strictly more than the number of iterations, but still of a comparable size. There may also be differences in how the radius of the trust region is calculated and updated between these two implementations. 
        
    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    n : int, optional
        Number of particles. The default is 96.
    K : int, optional
        Number of function calls allowed in TRR. Note that this uses the function calls reported by scipy, similar in scale to the number of iterations of TRR used in Loewe et al 2016. This is not the true number of function calls to the benchmarker as this appears to not include finite difference gradient calculations. The default is 5.
    Lmax : int, optional
        Maximum number of iterations. The default is 250.
    phi1 : float, optional
        Scale of the acceleration towards a particles best positions. The default is 2.05.
    phi2 : float, optional
        Scale of the acceleration towards the best point seen across all particles. The default is 2.05.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    
    class particle:
        def __init__(self, n_param):
            self.velocity = np.zeros(bm.n_parameters())
            self.position = np.random.rand(n_param)
            self.bestCost = np.inf #Best cost of this particle
            self.bestPosition = np.copy(self.position) #Position of best cost for this particle
            self.currentCost = None
        
        def set_cost(self, cost):
            self.currentCost = cost
            if cost < self.bestCost:
                self.bestCost = cost
                self.bestPosition = np.copy(self.position)
    
    def cost_func(x):
        return bm.cost(transform(x))
    
    def signed_error(x):
        return bm.signed_error(transform(x))
    
    def evaluate(x):
        return bm.evaluate(transform(x))
    
    def transform(x):
        if bm._bounded:
            xTrans = bm.lb + x*(bm.ub-bm.lb) #Map x to [lb,ub]
        else:
            xTrans = x*2 #Map x from [0,1] to [0,2]
            if not bm._useScaleFactors:
                xTrans = xTrans*bm.defaultParams #Map to [0,2*default]
        return xTrans
    
    if (phi1 + phi2)**2-4*(phi1 + phi2)<0:
        print("Invalid constriction factor using specified values for phi1 and phi2. Using defaults of phi1=phi2=2.05 instead.")
        phi1 = 2.05
        phi2 = 2.05
    phi = phi1 + phi2
    constFactor = 2/(phi-2+np.sqrt(phi**2-4*phi))
    
    if debug:
        verbose = 1
    else:
        verbose = 0
    
    particleList = []
    for i in range(n):
        particleList.append(particle(bm.n_parameters()))
    
    Gcost = [np.inf]*Lmax #Best cost ever
    Gpos = [None]*Lmax #Position of best cost ever
    for L in range(Lmax):
        if L > 0:
            Gcost[L] = Gcost[L-1]
            Gpos[L] = Gpos[L-1]
        
        if debug:
            print('-------------')
            print("Begginning population: "+str(L))
            print("Best cost so far: "+str(Gcost[L]))
            print("Found at position: "+str(Gpos[L]))
        
        #Find best positions, both globally and locally
        for p in particleList:
            cost = cost_func(p.position)
            p.set_cost(cost)
            if cost < Gcost[L]:
                Gcost[L] = cost
                Gpos[L] = np.copy(p.position)
        
        #Update velocities
        for p in particleList:
            localAcc = phi1*np.random.rand()*(p.bestPosition-p.position)
            globalAcc = phi2*np.random.rand()*(Gpos[L]-p.position)
            p.velocity = constFactor*(p.velocity + localAcc + globalAcc)
        if debug:
            print("Velocities renewed")
        
        #Enforce bounds
        for p in particleList:
            p.position += p.velocity
            for i in range(bm.n_parameters()):
                if p.position[i] < 0:
                    p.position[i] = np.random.rand()/4
                elif p.position[i] > 1:
                    p.position[i] = 1-np.random.rand()/4
        
        #Iterations of TRR
        if debug:
            print("Begginning TRR")
        bounds = ([0]*bm.n_parameters(),[1]*bm.n_parameters())
        for p in particleList:
            out = scipy.optimize.least_squares(signed_error, p.position, method='trf', diff_step=1e-3, max_nfev = 2*K, bounds = bounds, verbose=verbose)
            p.velocity = out.x - p.position
            p.position = out.x
            
        if debug:
            print("Positions renewed")
            print("Finished population: "+str(L))
            print("Best cost so far: "+str(Gcost[L]))
            print("Found at position: "+str(Gpos[L]))

    bm.evaluate(transform(Gpos[L]))
    return transform(Gpos[L])

if __name__ == '__main__':
    bm = ionbench.problems.staircase.HH_Benchmarker()
    run(bm,debug=True)

def get_approach():
    """
    The approach in Loewe et al uses a bounded search space with no transforms. The bounds are the same as for the sampler.

    Returns
    -------
    app : approach
        The approach used in Loewe et al 2016.

    """
    app = ionbench.approach.Loewe2016()
    return app



def multistart(opt, bm, initParams, filename, **kwargs):
    for i in range(len(initParams)):
        out = opt(bm, x0=initParams[i], **kwargs)
        bm.tracker.save(filename+'_run'+str(i)+'.pickle')
        print(out)
        bm.reset()



import ionbench
import myokit
import myokit.lib.markov as markov
import os
import numpy as np
import csv
import warnings

class ina(ionbench.benchmarker.Benchmarker):
    """
    The Moreno 2016 INa benchmarker. 
    
    The benchmarker uses the model from Moreno et al 2016 with a step protocol used to calculated summary curves which are then used for fitting. 
    
    Its parameters are specified as reported in Moreno et al 2016 with the true parameters being the same as the default and the center of the sampling distribution. 
    """
    def __init__(self):
        print('Initialising Moreno 2016 INa benchmark')
        self._name = "moreno2016.ina"
        self.model = myokit.load_model(os.path.join(ionbench.DATA_DIR, 'moreno2016', 'moreno2016.mmt'))
        self._outputName = 'ina.INa'
        self._paramContainer = 'ina'
        self.paramSpaceWidth = 5 #5, 10, or 25
        self.defaultParams = np.array([7.6178e-3, 3.2764e1, 5.8871e-1, 1.5422e-1, 2.5898, 8.5072, 1.3760e-3, 2.888, 3.2459e-5, 9.5951, 1.3771, 2.1126e1, 1.1086e1, 4.3725e1, 4.1476e-2, 2.0802e-2])
        self._rateFunctions = [(lambda p,V: 1/(p[0]*np.exp(-V/p[1])), 'negative'), (lambda p,V: p[2]/(p[0]*np.exp(-V/p[1])), 'negative'), (lambda p,V: p[3]/(p[0]*np.exp(-V/p[1])), 'negative'), (lambda p,V: 1/(p[4]*np.exp(V/p[5])), 'positive'), (lambda p,V: p[6]/(p[4]*np.exp(V/p[5])), 'positive'), (lambda p,V: p[7]/(p[4]*np.exp(V/p[5])), 'positive'), (lambda p,V: p[8]*np.exp(-V/p[9]), 'negative'), (lambda p,V: p[10]*np.exp(V/p[11]), 'positive'), (lambda p,V: p[12]*np.exp(V/p[13]), 'negative'), (lambda p,V: p[3]/(p[0]*np.exp(-V/p[1]))*p[12]*np.exp(V/p[13])*p[8]*np.exp(-V/p[9])/(p[7]/(p[4]*np.exp(V/p[5]))*p[10]*np.exp(V/p[11])), 'positive'), (lambda p,V: p[14]*p[12]*np.exp(V/p[13]), 'positive'), (lambda p,V: p[15]*p[8]*np.exp(-V/p[9]), 'negative')] #Used for rate bounds
        self.standardLogTransform = [True, False, True, True]*2+ [True, False]*3 + [True]*2
        self._trueParams = np.copy(self.defaultParams)
        self.load_data(dataPath = os.path.join(ionbench.DATA_DIR, 'moreno2016', 'ina.csv'))
        self._analyticalModel = markov.LinearModel(model = self.model, states = ['ina.'+s for s in ['ic3','ic2','if','c3','c2','c1','o','is']], parameters = [self._paramContainer+'.p'+str(i+1) for i in range(self.n_parameters())], current = self._outputName, vm = 'membrane.V')
        self.sim = myokit.Simulation(self.model, protocol=self.add_protocol())
        self.sim.pre(500) #Prepace for 500ms
        super().__init__()
        print('Benchmarker initialised')
    
    def sample(self, n=1):
        """
        Sample parameters for the Moreno 2016 problems. The sampling width can be changed by setting bm.paramSpaceWidth. The values used in Moreno et al 2016 are 5, 10, and 25. The default used is 5.

        Parameters
        ----------
        n : int, optional
            Number of parameter vectors to sample. The default is 1.

        Returns
        -------
        params : list
            If n=1, then params is the vector of parameters. Otherwise, params is a list containing n parameter vectors.

        """
        params = [None]*n
        for i in range(n):
            param = [None]*self.n_parameters()
            for j in range(self.n_parameters()):
                param[j] = self.defaultParams[j] * np.random.uniform(1-self.paramSpaceWidth/100,1+self.paramSpaceWidth/100)
            params[i] = self.input_parameter_space(param)
        if n==1:
            return params[0]
        else:
            return params
    
    def add_protocol(self):
        """
        Add the protocol from Moreno et al 2016. This is a series of steps, used for forming the summary statistic curves. This also sets the timepoints that need to be recorded to evaluate these summary statistics.

        Returns
        -------
        p : myokit.protocol
            A myokit.protocol for the voltage clamp protocol from Loewe et al 2016.

        """
        #Setup
        measurementWindows = []
        gap = 5000
        newProtocol = myokit.Protocol()
        
        #Protocol 1 - Measure peak current at -10mV after holding at voltages between -120mV and -40mV
        #Track start times
        protocolStartTimes = [0]
        #Create protocol
        protocol = myokit.pacing.steptrain_linear(vstart=-120, vend=-30, dv=10, vhold=-10, tpre=0, tstep=gap, tpost=gap)
        #Add windows to measure at
        for e in protocol.events():
            if e.level()==-10:
                measurementWindows.append([e.start(),e.stop(),'ssi'])
        #Add protocol to full protocol
        for e in protocol.events():
            newProtocol.schedule(level=e.level(), start=e.start(), duration=e.duration())
        #Add barrier to separate effects from different protocols
        newProtocol.schedule(level=-100, start=newProtocol.characteristic_time(), duration=gap)
        
        #Protocol 2 - Measure steady state current at varying voltages between -75mV and 20mV
        #Track start times
        protocolStartTimes.append(newProtocol.characteristic_time())
        #Create protocol
        protocol = myokit.pacing.steptrain_linear(vstart=-75, vend=25, dv=5, vhold=-100, tpre=gap, tstep=gap, tpost=0)
        #Add windows to measure at
        offset = newProtocol.characteristic_time()
        for e in protocol.events():
            if e.level()!=-100:
                measurementWindows.append([e.stop()+offset-0.01,'act']) #Just before transition
        #Add protocol to full protocol
        for e in protocol.events():
            newProtocol.schedule(level=e.level(), start=offset+e.start(), duration=e.duration())
        #Add barrier to separate effects from different protocols
        newProtocol.schedule(level=-100, start=newProtocol.characteristic_time(), duration=gap)
        
        #Protocol 3 - Ratio of max current at steps to -10mV with varying length gaps between
        #Track start times
        protocolStartTimes.append(newProtocol.characteristic_time())
        #Create protocol
        dt = [0.1, 0.5, 1, 3, 6, 9, 12, 15, 21, 30, 60, 90, 120, 210, 300, 450, 600, 900, 1500, 3000, 6000]
        vsteps = [-100, -10, -100, -10]*len(dt)+[-100]
        times = []
        for i in range(len(dt)):
            times.append(gap)
            times.append(100)
            times.append(dt[i])
            times.append(25)
        times.append(gap)
        protocol = myokit.Protocol()
        for i in range(len(times)):
            protocol.add_step(vsteps[i],times[i])
        #Add windows to measure at
        offset = newProtocol.characteristic_time()
        for e in protocol.events():
            if e.level()==-10:
                measurementWindows.append([e.start()+offset,e.stop()+offset,'rfi'])
        #Add protocol to full protocol
        for e in protocol.events():
            newProtocol.schedule(level=e.level(), start=offset+e.start(), duration=e.duration())
        #Add barrier to separate effects from different protocols
        newProtocol.schedule(level=-100, start=newProtocol.characteristic_time(), duration=gap)
        
        #Protocol 4 - Ratio of first and last max currents at 300 steps to -10mV with varying length gaps between
        #Track start times
        protocolStartTimes.append(newProtocol.characteristic_time())
        #Create protocol
        dt = [0.5, 1, 3, 9, 30, 90, 300, 900, 3000, 9000]
        vsteps = []
        times = []
        for i in range(len(dt)):
            vsteps.append(-100)
            times.append(gap)
            vsteps += [-10, -100]*300
            times += [25,dt[i]]*300
        times.append(gap)
        vsteps.append(-100)
        protocol = myokit.Protocol()
        for i in range(len(times)):
            protocol.add_step(vsteps[i],times[i])
        #Add windows to measure at
        offset = newProtocol.characteristic_time()
        tmpOffset = offset
        for i in range(len(dt)):
            measurementWindows.append([tmpOffset+gap,tmpOffset+gap+25,'rudb'])
            measurementWindows.append([tmpOffset+gap+(25+dt[i])*299,tmpOffset+gap+(25+dt[i])*299+25,'rudb'])
            tmpOffset = tmpOffset+gap+(25+dt[i])*300
        #Add protocol to full protocol
        for e in protocol.events():
            newProtocol.schedule(level=e.level(), start=offset+e.start(), duration=e.duration())
        #Add barrier to separate effects from different protocols
        newProtocol.schedule(level=-100, start=newProtocol.characteristic_time(), duration=gap)
        
        #Protocol 5 - Time to 50% of max current after step to between -20mV and 20mV
        #Track start times
        protocolStartTimes.append(newProtocol.characteristic_time())
        #Create protocol
        protocol = myokit.pacing.steptrain_linear(vstart=-20, vend=25, dv=5, vhold=-100, tpre=gap, tstep=gap, tpost=0)
        #Add windows to measure at
        offset = newProtocol.characteristic_time()
        for e in protocol.events():
            if e.level()!=-100:
                measurementWindows.append([e.start()+offset,e.stop()+offset,'tau'])
        #Add protocol to full protocol
        for e in protocol.events():
            newProtocol.schedule(level=e.level(), start=offset+e.start(), duration=e.duration())
        
        #Track total time
        protocolStartTimes.append(newProtocol.characteristic_time())
        
        #Store measurement windows
        self.tmax = newProtocol.characteristic_time()
        
        self._logTimes = []
        self._ssiBounds = []
        self._actBounds = []
        self._rfiBounds = []
        self._rudbBounds = []
        self._tauBounds = []
        
        for i in measurementWindows:
            if i[-1] == 'act':
                self._logTimes.append(i[0])
                self._actBounds.append(len(self._logTimes))
            else:
                lb = len(self._logTimes)
                self._logTimes += list(np.arange(i[0],i[1],0.005))
                ub = len(self._logTimes)
                if i[-1] == 'ssi':
                    self._ssiBounds.append([lb,ub])
                elif i[-1] == 'rfi':
                    self._rfiBounds.append([lb,ub])
                elif i[-1] == 'rudb':
                    self._rudbBounds.append([lb,ub])
                elif i[-1] == 'tau':
                    self._tauBounds.append([lb,ub])
        self._logTimes = np.array(self._logTimes)
        return newProtocol
    
    def solve_model(self, times, continueOnError = True):
        """
        Replaces the Benchmarker solve_model to call a special Moreno 2016 method (run_moreno()) which handles the summary curve calculations. The output is a vector of points on the summary curves.

        Parameters
        ----------
        times : list or numpy array
            Unneccessary for Moreno 2016. Only kept in since it will be passed in as an input by the main Benchmarker methods.
        continueOnError : bool, optional
            If continueOnError is True, any errors that occur during solving the model will be ignored and an infinite output will be given. The default is True.

        Returns
        -------
        modelOutput : list
            A vector of points on summary curves.

        """
        if continueOnError:
            try:
                return self.run_moreno()
            except:
                warnings.warn("Failed to solve model. Will report infinite output in the hope of continuing the run.")
                return np.array([np.inf]*69)
        else:
            return self.run_moreno()
    
    def cost(self, parameters, incrementSolveCounter = True):
        """
        Find the RMSE cost between the model solved using the inputted parameters and the data. For Moreno, this method weights the contributions of the summary curves by the reciprical of the number of points in each summary curve. 

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.
        incrementSolveCounter : bool, optional
            If False, it disables the solve counter tracker. This never needs to be set to False by a user. This is only required by the evaluate() method. The default is True.

        Returns
        -------
        cost : float
            The RMSE cost of the parameters.
        
        """
        weights = [1/9]*9+[1/20]*20+[1/21]*21+[1/10]*10+[1/9]*9
        testOutput = np.array(self.simulate(parameters, np.arange(0, self.tmax), incrementSolveCounter = incrementSolveCounter))
        cost = np.sqrt(np.average((testOutput-self.data)**2, weights=weights))
        return cost
    
    def run_moreno(self):
        """
        Runs the model to generate the Moreno et al 2016 summary curves. The points on these summary curves are then returned.

        Returns
        -------
        modelOutput : list
            A vector of points on summary curves.

        """
        
        # Run a simulation
        log = self.sim.run(self.tmax+1, log_times = self._logTimes, log = [self._outputName])
        inaOut = -np.array(log[self._outputName])
        
        ssi = [-1]*9
        for i in range(len(ssi)):
            ssi[i] = max(inaOut[self._ssiBounds[i][0]:self._ssiBounds[i][1]])
        ssi = [a/max(ssi) for a in ssi]
        
        act = [-1]*20
        for i in range(len(act)):
            act[i] = inaOut[self._actBounds[i]]
        act = [a/max(act) for a in act]
        
        rfi = [-1]*21
        for i in range(len(rfi)):
            firstPulse = max(inaOut[self._rfiBounds[2*i][0]:self._rfiBounds[2*i][1]])
            secondPulse = max(inaOut[self._rfiBounds[2*i+1][0]:self._rfiBounds[2*i+1][1]])
            rfi[i] = secondPulse/firstPulse
        
        rudb = [-1]*10
        for i in range(len(rudb)):
            firstPulse = max(inaOut[self._rudbBounds[2*i][0]:self._rudbBounds[2*i][1]])
            lastPulse = max(inaOut[self._rudbBounds[2*i+1][0]:self._rudbBounds[2*i+1][1]])
            rudb[i] = lastPulse/firstPulse
        
        tau = [-1]*9
        for i in range(len(tau)):
            peak = max(inaOut[self._tauBounds[i][0]:self._tauBounds[i][1]])
            reachedPeak = False
            for j in range(1000000):
                current = inaOut[self._tauBounds[i][0]+j]
                if current == peak:
                    reachedPeak = True
                    timeToPeak = self._logTimes[self._tauBounds[i][0]+j]
                if reachedPeak and abs(current)<=abs(peak/2):
                    tau[i] = self._logTimes[self._tauBounds[i][0]+j]-timeToPeak
                    break
        return np.array(ssi+act+rfi+rudb+tau)

def generate_data():
    """
    Generate the data files for the Moreno 2016 benchmarker problems. The true parameters are the same as the deafult for these benchmark problems.

    Parameters
    ----------
    None.

    Returns
    -------
    None.

    """
    bm = ina()
    out = bm.solve_model(np.arange(bm.tmax), continueOnError = False)
    with open(os.path.join(ionbench.DATA_DIR, 'moreno2016', 'ina.csv'), 'w', newline = '') as csvfile:
        writer = csv.writer(csvfile, delimiter = ',')
        writer.writerows(map(lambda x: [x], out))



import ionbench
import numpy as np
import matplotlib.pyplot as plt
import pickle

class ProfileManager():
    def __init__(self, bm, fixedParam, fixedValue):
        """
        A middle-man for handling profile likelihood plots. This class acts as a wrapper for the benchmarker, such that the benchmarker always sees the full parameter vector, while the optimiser sees a reduced vector with one parameter missing.

        Parameters
        ----------
        bm : benchmarker
            A benchmarker to act as the wrapper for.
        fixedParam : int
            The index of the parameter to be fixed.
        fixedValue : float
            The value at which to fix the parameter.

        Returns
        -------
        None.

        """
        self.bm = bm
        self.fixedParam = fixedParam
        self.fixedValue = fixedValue
        self.MLE = bm._trueParams
        self._bounded = False
    
    def n_parameters(self):
        """
        ProfileManager wrapper for the benchmarker's .n_parameters() method. Reduces the parameter count by 1 to account for the fixed parameter.
        """
        return self.bm.n_parameters()-1
    
    def cost(self, params):
        """
        ProfileManager wrapper for the benchmarker's .cost() method. Inserts the fixed parameter before evaluating.
        """
        return self.bm.cost(self.set_params(params))
    
    def signed_error(self, params):
        """
        ProfileManager wrapper for the benchmarker's .signed_error() method. Inserts the fixed parameter before evaluating.
        """
        return self.bm.signed_error(self.set_params(params))
    
    def set_params(self, params):
        """
        Inserts the fixed parameter into the inputted parameter vector.
        """
        return np.insert(params, self.fixedParam, self.fixedValue)
    
    def sample(self):
        """
        ProfileManager wrapper for the benchmarker's .sample() method. Since sampling is used to define the initial points, this method returns the true parameters with the fixed parameter removed.
        """
        return np.delete(self.MLE, self.fixedParam)
    
    def evaluate(self, x):
        """
        ProfileManager wrapper for the benchmarker's .evaluate() method. This method does nothing, but is neccessary since optimisers will attempt to call it.
        """
        pass

def run(bm, variations, plot = True, filename = ''):
    """
    Generate a profile likelihood style plot, reporting the fitted cost as a function of each fixed parameter.

    Parameters
    ----------
    bm : benchmarker
        A benchmarker function to explore.
    variations : list
        A list of parameter variations to apply for each parameter. It should be a list of length bm.n_parameters(), where the ith element is a list of variations to use to fix parameter i and fit the remaining parameters. 
    plot : bool, optional
        Whether or not to plot the likelihood and give plots of the current at the start and end points to see if the variability is significant. The default is True.
    filename : string, optional
        A name to use to save the cost data. It will pickle the cost and variations and save them under [filename]_param[i].pickle. The default is '', in which case no data will be saved.

    Returns
    -------
    None.

    """
    bm.plotter = False
    for i in range(bm.n_parameters()):
        print('Working on parameter '+str(i))
        costs = np.zeros(len(variations[i]))
        for j in range(len(variations[i])):
            bm.reset()
            var = variations[i][j]
            pm = ProfileManager(bm, i, bm._trueParams[i]*var)
            if var == 1:
                costs[j] = bm.cost(bm._trueParams)
                out = pm.sample()
            else:
                try:
                    out = ionbench.optimisers.scipy_optimisers.lm_scipy.run(pm)
                except:
                    out = pm.sample()
                costs[j] = pm.cost(out)
            print('Variation: '+str(var))
            print('Cost found: '+str(costs[j]))
            if j==0 or j==len(variations[i])-1 and plot:
                bm.sim.reset()
                bm.set_params(pm.set_params(out))
                curr1 = bm.solve_model(np.arange(bm.tmax), continueOnError = True)
                bm.sim.reset()
                bm.set_params(bm._trueParams)
                curr2 = bm.solve_model(np.arange(bm.tmax), continueOnError = True)
                plt.figure()
                plt.plot(np.arange(bm.tmax),curr1,'k')
                plt.plot(np.arange(bm.tmax),curr2,'b')
                plt.ylabel('Current')
                plt.xlabel('Time (ms)')
                plt.legend(['Fitted','Data'])
                plt.show()
        if plot:
            plt.figure()
            plt.plot(variations[i], costs)
            plt.xlabel('Parameter '+str(i))
            plt.title('Profile Likelihood')
            plt.ylabel('Cost')
            plt.show()
        if not filename == '':
            data = (variations[i], costs)
            with open(filename+'_param'+str(i)+'.pickle', 'wb') as f:
                pickle.dump(data,f)

def plot_profile_likelihood(modelType, numberToPlot, ymax=0):
    for i in range(numberToPlot):
        with open(modelType+'_param'+str(i)+'.pickle', 'rb') as f:
            variations, costs = pickle.load(f)
        plt.figure()
        plt.plot(variations[i], costs)
        if not ymax == 0:
            plt.ylim(0,ymax)
        plt.title('Profile likelihood: '+modelType)
        plt.xlabel('Factor for parameter '+str(i))
        plt.ylabel('Cost')
        plt.show()



import pints
import numpy as np

class Model(pints.ForwardModel):
    """
    A Pints forwards model containing a benchmarker class.
    """
    def __init__(self, bm):
        """
        Initialise a Pints forward model with a benchmarker, linking up the n_parameters() and simulate() methods.

        Parameters
        ----------
        bm : benchmarker
            A test problem benchmarker.

        Returns
        -------
        None.

        """
        self.bm = bm
        
    def n_parameters(self):
        """
        Returns the number of parameters in the model

        Returns
        -------
        n_parameters : int
            Number of parameters in the model.

        """
        return self.bm.n_parameters()
    
    def simulate(self, parameters, times):
        """
        Simulates the model and returns the model output.

        Parameters
        ----------
        parameters : list
            A list of parameter values, length n_parameters().
        times : list
            A list of times at which to return model output.

        Returns
        -------
        out : list
            Model output, typically a current trace.

        """
        # Reset the simulation
        return self.bm.simulate(parameters, times)

class AdvancedBoundaries(pints.Boundaries):
    """
    Pints boundaries to apply to the parameters and the rates. 
    """
    def __init__(self, bm, vHigh = 40, vLow = -120):
        """
        Build a Pints boundary object to apply parameter and rate bounds.

        Parameters
        ----------
        bm : benchmarker
            A test problem benchmarker.
        vHigh : float, optional
            The high voltage to use for the rate bounds. The default is 40.
        vLow : float, optional
            The low voltage to use for the rate bounds. The default is -120.

        Returns
        -------
        None.

        """
        self.bm = bm
        self.km_min = 1.67e-5
        self.km_max = 1e3
        self.vLow = vLow
        self.vHigh = vHigh
    
    def n_parameters(self):
        """
        Returns the number of parameters in the model.
        """
        return self.bm.n_parameters()
    
    def check(self, parameters):
        """
        Check inputted parameters against the parameter and rate bounds.

        Parameters
        ----------
        parameters : list
            Inputted parameter to check, in the input parameter space.

        Returns
        -------
        paramsInsideBounds : bool
            True if the parameters are inside the bound. False if the parameters are outside of the bounds.

        """
        parameters = self.bm.original_parameter_space(parameters)
        
        # Check parameter boundaries
        if np.any(parameters <= self.bm.lb) or np.any(parameters >= self.bm.ub):
            return False
        
        #Check rate boundaries
        for rateTuple in self.bm._rateFunctions:
            rateFunc = rateTuple[0]
            rateType = rateTuple[1]
            if rateType == 'positive':
                #check kHigh is in bounds
                k = rateFunc(parameters, self.vHigh)
            elif rateType == 'negative':
                #check kLow is in bounds
                k = rateFunc(parameters, self.vLow)
            elif rateType == 'independent':
                #check rate in bounds
                k = rateFunc(parameters, 0) #Voltge doesn't matter
            else:
                print("Error in bm._rateFunctions. Doesn't contain 'positive', 'negative', or 'independent' in atleast one position. Check for typos.")
                k=0
            if k < self.km_min or k > self.km_max:
                return False
        
        # All tests passed!
        return True



from ionbench.optimisers.scipy_optimisers import lm_scipy
from ionbench.optimisers.scipy_optimisers import nelderMead_scipy
from ionbench.optimisers.scipy_optimisers import powell_scipy
from ionbench.optimisers.scipy_optimisers import trustRegionReflective_scipy



import ionbench.problems.staircase
import scipy.optimize
import numpy as np

def run(bm, x0 = [], xtol = 1e-4, ftol = 1e-4, maxiter = 5000, maxfev = 20000):
    """
    Runs Powell's Simplex optimiser from Scipy. Bounds are automatically loaded from the benchmarker if present.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is [], in which case a randomly sampled parameter vector is retrieved from bm.sample().
    xtol : float, optional
        Tolerance in parameters. Used as a termination criterion. The default is 1e-4.
    ftol : float, optional
        Tolerance in cost. Used as a termination criterion. The default is 1e-4.
    maxiter : int, optional
        Maximum number of iterations of Powell's Simplex to use. The default is 5000.
    maxfev : int, optional
        Maximum number of cost function evaluations. The default is 20000.

    Returns
    -------
    xbest : list
        The best parameters identified by Powell's Simplex.

    """
    if len(x0)==0:
        x0 = bm.sample()
    
    if bm._bounded:
        lb = bm.lb[:] #Generate copy
        ub = bm.ub[:] #Generate copy
        for i in range(bm.n_parameters()):
            if lb[i] == np.inf:
                lb[i] = None
            if ub[i] == np.inf:
                ub[i] = None
        bounds = (lb,ub)
        
        out = scipy.optimize.minimize(bm.cost, x0, method='powell', options={'disp': True, 'xtol': xtol, 'ftol': ftol, 'maxiter': maxiter, 'maxfev': maxfev}, bounds = bounds)
    else:
        out = scipy.optimize.minimize(bm.cost, x0, method='powell', options={'disp': True, 'xtol': xtol, 'ftol': ftol, 'maxiter': maxiter, 'maxfev': maxfev})
        
    bm.evaluate(out.x)
    return out.x

if __name__ == '__main__':
    bm = ionbench.problems.staircase.HH_Benchmarker()
    run(bm)

def get_approach():
    """
    No approach specified. Will use an empty approach

    Returns
    -------
    app : approach
        Empty approach

    """
    app = ionbench.approach.Empty(name = 'powell_scipy')
    return app



import ionbench
from functools import lru_cache
import numpy as np

def run(bm, x0 = [], CrtStp = 2e-5, Stp = 1/100, RedFct = 1/4, maxfev = 100000, debug = False):
    """
    Runs the pattern search algorithm from Kohjitani et al 2022.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is [], in which case a randomly sampled parameter vector is retrieved from bm.sample().
    CrtStp : float, optional
        The minimum step size. If there are no improvements within CrtStp, the optimisation terminates. The default is 2e-5.
    Stp : float, optional
        Initial step size. The default is 1/100.
    RedFct : float, optional
        The reduction factor. If the center point is better than its neighbours, the step size is scaled by this reduction factor. The default is 1/4.
    maxfev : int, optional
        Maximum number of cost function evaluations. The default is 100000.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    @lru_cache(maxsize=None)
    def cost_func(x):
        return bm.cost(x)
    
    funcCounter = 0
    def explore(BP, Stp):
        foundImprovement = False
        NP = np.copy(BP)
        MSE = cost_func(tuple(BP)) #No real computation cost from this thanks to the cache
        for i in range(bm.n_parameters()): 
            home = BP[i]
            NP[i] = home + Stp#Positive Step
            MSEp = cost_func(tuple(NP))#Positive MSE
            NP[i] = home - Stp#Negative Step
            MSEn = cost_func(tuple(NP))#Negative MSE
            minMSE = min(MSEp, MSEn)#MSE in best direction (postive or negative Stp)
            if minMSE < MSE: #If improvement found
                if MSEp < MSEn: #If positive step is better
                    NP[i] = home + Stp#Take positive step
                else:#If negative step is better
                    NP[i] = home - Stp#Take negative step
                MSE = minMSE#Either way, record new MSE
                foundImprovement = True
            else:#If no improvement
                NP[i] = home#Restore center point
        return foundImprovement, NP
    
    if len(x0)==0:
        x0 = bm.sample()
    
    BP = x0 #Set initial base point
    funcCounter += 1
    while Stp > CrtStp: #Stop when step size is sufficiently small
        #Explore neighbouring points
        if debug:
            print("------------")
            print("Current step size:"+str(Stp))
            print("Cost: "+str(cost_func(tuple(BP))))
        improvementFound, NP = explore(BP, Stp) #Explore neighbouring points
        funcCounter += 4
        while improvementFound:
            if debug:
                print("Improvement Found? "+str(improvementFound))
            BP = np.copy(NP) #Move to new improved point
            improvementFound, NP = explore(BP, Stp) #Explore neighbouring points
            funcCounter += 4
            if funcCounter > maxfev:
                print("Exceeded maximum number of function evaluations.")
                bm.evaluate(NP)
                return NP
        
        Stp = Stp * RedFct #Decrease step size if all neighbouring points are worse
    
    bm.evaluate(NP)
    return NP

if __name__ == '__main__':
    bm = ionbench.problems.staircase.HH_Benchmarker()
    run(bm, debug = True)

def get_approach():
    """
    The approach in Kojitani et al 2022 uses scaling factors. They have bounds on the sampling but I don't think they use these in the optimisation.

    Returns
    -------
    app : approach
        The approach used in Kohjitani et al 2022.

    """
    app = ionbench.approach.Kohjitani2022()
    return app



import ionbench
import myokit
import myokit.lib.hh as hh
import os
import numpy as np
import csv

class loewe2016_Benchmarker(ionbench.benchmarker.Benchmarker):
    def __init__(self):
        self._trueParams = np.copy(self.defaultParams)
        self.paramSpaceWidth = 1 #1 for narrow, 2 for wide
        self.standardLogTransform = [not i for i in self.additiveParams]
        super().__init__()
    
    def sample(self, n=1):
        """
        Sample parameters for the Loewe 2016 problems. By default the sampling using the narrow parameter space but this can be changed by setting benchmarker.paramSpaceWidth = 2 to use the wide parameter space.

        Parameters
        ----------
        n : int, optional
            Number of parameter vectors to sample. The default is 1.

        Returns
        -------
        params : list
            If n=1, then params is the vector of parameters. Otherwise, params is a list containing n parameter vectors.

        """
        params = [None]*n
        for i in range(n):
            param = [None]*self.n_parameters()
            for j in range(self.n_parameters()):
                if self.additiveParams[j]:
                    param[j] = self.defaultParams[j] + np.random.uniform(-60*self.paramSpaceWidth,60*self.paramSpaceWidth)
                else:
                    param[j] = self.defaultParams[j]*10**np.random.uniform(-1*self.paramSpaceWidth,1*self.paramSpaceWidth) #Log uniform distribution
            params[i] = self.input_parameter_space(param) #Generates a copy
        if n==1:
            return params[0]
        else:
            return params
        
    def add_protocol(self):
        """
        Add the protocol from Loewe et al 2016. This protocol consists of 20ms at -80mV, a variable height step between 50mV and -70mV (in steps of 10mV) for 400ms and then 400ms at -110mV before repeating with a new height for the central step (in order of decreasing voltage).

        Returns
        -------
        p : myokit.protocol
            A myokit.protocol for the voltage clamp protocol from Loewe et al 2016.

        """
        p = myokit.Protocol()
        vsteps = []
        for i in range(13):
            vsteps.append(-80)
            vsteps.append(50-i*10)
            vsteps.append(-110)
        durations = [20,400,400]*13
        for i in range(len(vsteps)):
            p.add_step(vsteps[i],durations[i])
        self.tmax = sum(durations)
        return p
        

class ikr(loewe2016_Benchmarker):
    """
    The Loewe 2016 IKr benchmarker. 
    
    The benchmarker uses the Courtemanche 1998 IKr model with a simple step protocol. 
    
    Its parameters are specified as reported in Loewe et al 2016 with the true parameters being the same as the default and the center of the sampling distribution. 
    """
    def __init__(self):
        print('Initialising Loewe 2016 IKr benchmark')
        self._name = "loewe2016.ikr"
        self.model = myokit.load_model(os.path.join(ionbench.DATA_DIR, 'loewe2016', 'courtemanche-1998-ikr.mmt'))
        self._outputName = 'ikr.IKr'
        self._paramContainer = 'ikr'
        self.defaultParams = np.array([3e-4, 14.1, 5, 3.3328, 5.1237, 1, 14.1, 6.5, 15, 22.4, 0.029411765, 138.994])
        self.additiveParams = [False, True, False, True, False, False, True, False, True, False, False, False]
        self._rateFunctions = [(lambda p,V:p[0]*(V+p[1])/(1-np.exp((V+p[1])/(-p[2]))), 'positive'), (lambda p,V:7.3898e-5*(V+p[3])/(np.exp((V+p[3])/p[4])-1), 'negative')] #Used for rate bounds
        self.load_data(dataPath = os.path.join(ionbench.DATA_DIR, 'loewe2016', 'ikr.csv'))
        self._analyticalModel = hh.HHModel(model = self.model, states = ['ikr.xr'], parameters = [self._paramContainer+'.p'+str(i+1) for i in range(self.n_parameters())], current = self._outputName, vm = 'membrane.V')
        self.sim = hh.AnalyticalSimulation(self._analyticalModel, protocol=self.add_protocol())
        self.sim.pre(500) #Prepace for 500ms
        super().__init__()
        print('Benchmarker initialised')

class ikur(loewe2016_Benchmarker):
    """
    The Loewe 2016 IKur benchmarker. 
    
    The benchmarker uses the Courtemanche 1998 IKur model with a simple step protocol. 
    
    Its parameters are specified as reported in Loewe et al 2016 with the true parameters being the same as the default and the center of the sampling distribution. 
    """
    def __init__(self):
        print('Initialising Loewe 2016 IKur benchmark')
        self._name = "loewe2016.ikur"
        self.model = myokit.load_model(os.path.join(ionbench.DATA_DIR, 'loewe2016', 'courtemanche-1998-ikur.mmt'))
        self._outputName = 'ikur.IKur'
        self._paramContainer = 'ikur'
        self.defaultParams = np.array([0.65, 10, 8.5, 30, 59, 2.5, 82, 17, 30.3, 9.6, 3, 1, 21, 185, 28, 158, 16, 99.45, 27.48, 3, 0.005, 0.05, 15, 13, 138.994])
        self.additiveParams = [False, True, False, True, False, True, True, False, True, False, False, False, True, True, False, True, True, True, False, False, True, False, True, False, False]
        self._rateFunctions = [(lambda p,V: p[0]/(np.exp((V+p[1])/-p[2])+np.exp((V-p[3])/-p[4])), 'positive'), (lambda p,V: 0.65/(p[5]+np.exp((V+p[6])/p[7])), 'negative'), (lambda p,V: p[11]/(p[12]+np.exp((V-p[13])/-p[14])), 'positive'), (lambda p,V: np.exp((V-p[15])/-p[16]), 'negative')] #Used for rate bounds
        self.load_data(dataPath = os.path.join(ionbench.DATA_DIR, 'loewe2016', 'ikur.csv'))
        self._analyticalModel = myokit.lib.hh.HHModel(model = self.model, states = ['ikur.ua', 'ikur.ui'], parameters = [self._paramContainer+'.p'+str(i+1) for i in range(self.n_parameters())], current = self._outputName, vm = 'membrane.V')
        self.sim = myokit.lib.hh.AnalyticalSimulation(self._analyticalModel, protocol=self.add_protocol())
        self.sim.pre(500) #Prepace for 500ms
        super().__init__()
        print('Benchmarker initialised')

def generate_data(modelType):
    """
    Generate the data files for the Loewe 2016 benchmarker problems. The true parameters are the same as the deafult for these benchmark problems.

    Parameters
    ----------
    modelType : string
        'ikr' to generate the data for the IKr benchmark problem. 'ikur' to generate the data for the IKur benchmark problem.

    Returns
    -------
    None.

    """
    modelType = modelType.lower()
    if modelType == 'ikr':
        bm = ikr()
    elif modelType == 'ikur':
        bm = ikur()
    out = bm.simulate(bm.defaultParams, np.arange(bm.tmax), continueOnError = False)
    with open(os.path.join(ionbench.DATA_DIR, 'loewe2016', modelType.lower()+'.csv'), 'w', newline = '') as csvfile:
        writer = csv.writer(csvfile, delimiter = ',')
        writer.writerows(map(lambda x: [x], out))



import os
from ionbench import benchmarker
from ionbench import problems
from ionbench import optimisers
from ionbench import approach
from ionbench import uncertainty
from ionbench.multistart import *

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

DATA_DIR = os.path.join(ROOT_DIR,"data")

TEST_DIR = os.path.join(os.path.dirname(ROOT_DIR),"test")


import pints
import ionbench
import numpy as np
from ionbench.optimisers.pints_optimisers import classes_pints

def run(bm, x0 = [], maxIter=1000):
    """
    Runs XNES (Exponential Natural Evolution Strategy) from Pints using a benchmarker. 

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is [], in which case a randomly sampled parameter vector is retrieved from bm.sample().
    maxIter : int, optional
        Number of iterations of XNES to run. The default is 1000.

    Returns
    -------
    xbest : list
        The best parameters identified by XNES.

    """
    if len(x0)==0:
        x0 = bm.sample()
    model = classes_pints.Model(bm)
    problem = pints.SingleOutputProblem(model, np.arange(model.bm.tmax), model.bm.data)
    error = pints.RootMeanSquaredError(problem)
    
    # Create an optimisation controller
    opt = pints.OptimisationController(error, x0, method=pints.XNES)
    opt.set_max_iterations(maxIter)
    # Run the optimisation
    x, f = opt.run()
    
    model.bm.evaluate(x)
    return x

if __name__ == '__main__':
    bm = ionbench.problems.staircase.HH_Benchmarker()
    bm.log_transform([True, False]*4+[False])
    run(bm)

def get_approach():
    """
    No approach specified. Will use an empty approach

    Returns
    -------
    app : approach
        Empty approach

    """
    app = ionbench.approach.Empty(name = 'xnes_pints')
    return app



import numpy as np
import scipy
import ionbench
from functools import lru_cache
import copy

from pymoo.core.individual import Individual
from pymoo.core.problem import Problem
from pymoo.operators.crossover.sbx import SBX

def run(bm, nGens = 50, eta_cross = 10, eta_mut = 20, elitePercentage = 0.066, popSize = 50, debug = False):
    """
    Runs the genetic algorithm from Smirnov et al 2020.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    nGens : int, optional
        The number of generations to run the optimisation algorithm for. The default is 50.
    eta_cross : float, optional
        Crossover parameter. The default is 10.
    eta_mut : float, optional
        Mutation parameter. The default is 20.
    elitePercentage : float, optional
        The percentage of the population that are considered elites to move into the next generation. This will be multiplied by popSize and then rounded to the nearest integer. The default is 0.066.
    popSize : int, optional
        The size of the population in each generation. The default is 50.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    class individual():
        def __init__(self):
            self.x = bm.sample()
            self.cost = None
        def find_cost(self):
            self.cost = cost_func(tuple(self.x))
    
    @lru_cache(maxsize=None)
    def cost_func(x):
        return bm.cost(x)
    
    eliteCount = int(np.round(popSize*elitePercentage))
    pop = [None]*popSize
    for i in range(popSize):
        pop[i] = individual()
        pop[i].find_cost()
    
    for gen in range(nGens):
        costVec = [0]*popSize
        for i in range(popSize):
            costVec[i] = pop[i].cost
        eliteIndices = np.argsort(costVec)[:eliteCount]
        elites = [None]*eliteCount
        for i in range(eliteCount):
            elites[i] = copy.deepcopy(pop[eliteIndices[i]])
        if debug:
            print("------------")
            print("Gen "+str(gen))
            print("Best cost: "+str(min(costVec)))
            print("Average cost: "+str(np.mean(costVec)))
        #Tournement selection
        newPop = []
        for j in range(2):
            perm = np.random.permutation(popSize)
            for i in range(popSize//2):
                if pop[perm[2*i]].cost > pop[perm[2*i+1]].cost:
                    newPop.append(copy.deepcopy(pop[perm[2*i]]))
                else:
                    newPop.append(copy.deepcopy(pop[perm[2*i+1]]))
        pop = newPop #Population of parents
        #Crossover SBX
        newPop = []
        problem = Problem(n_var=bm.n_parameters(), xl=0.0, xu=2.0)
        for i in range(popSize//2):
            a, b = Individual(X=np.array(pop[2*i].x)), Individual(X=np.array(pop[2*i+1].x))
    
            parents = [[a, b]]
            off = SBX(prob=0.9, eta=eta_cross).do(problem, parents) #What is prob vs prob_var
            Xp = off.get("X")
            newPop.append(individual())
            newPop[-1].x = Xp[0]
            newPop.append(individual())
            newPop[-1].x = Xp[1] #Can this be done in one line
        pop = newPop
        #Mutation
        for i in range(popSize):
            if np.random.rand()<0.9:
                direc = np.random.rand(bm.n_parameters())
                direc = direc/np.linalg.norm(direc)
                mag = scipy.stats.cauchy.rvs(loc=0, scale=0.18)
                pop[i].x += mag*direc
        if debug:
            print("Finishing gen "+str(gen))
        #Find costs
        for i in range(popSize):
            pop[i].find_cost()
        #Elitism
        costVec = [0]*popSize
        for i in range(popSize):
            costVec[i] = pop[i].cost
        eliteIndices = np.argsort(costVec)[-eliteCount:]
        for i in range(eliteCount):
            pop[eliteIndices[i]] = copy.deepcopy(elites[i])
    
    minCost = np.inf
    for i in range(popSize):
        if pop[i].cost < minCost:
            minCost = pop[i].cost
            elite = pop[i]
    bm.evaluate(elite.x)
    return elite.x


if __name__ == '__main__':
    bm = ionbench.problems.staircase.HH_Benchmarker()
    run(bm, debug=True)

def get_approach():
    """
    No approach settings given in Smirnov et al 2020. Will use an empty approach

    Returns
    -------
    app : approach
        Empty approach

    """
    app = ionbench.approach.Empty(name = 'Smirnov2020')
    return app



from ionbench.optimisers.pints_optimisers import cmaes_pints
from ionbench.optimisers.pints_optimisers import nelderMead_pints
from ionbench.optimisers.pints_optimisers import pso_pints
from ionbench.optimisers.pints_optimisers import snes_pints
from ionbench.optimisers.pints_optimisers import xnes_pints



from ionbench.problems import staircase
import spsa
#To do:
#Convergence needs to massively improve. An implementation of the original Spall 1992 algorithm might actually help

def run(bm, x0 = [], printFrequency=100, maxiter = 1000):
    """
    Runs SPSA (Simultaneous Perturbation Stochastic Approximation) from Pints using a benchmarker. 

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is [], in which case a randomly sampled parameter vector is retrieved from bm.sample().
    printFrequency : int, optional
        Number of iterations to wait between printing output
    maxIter : int, optional
        Number of iterations of SPSA to run per repeat. The default is 1000.

    Returns
    -------
    xbest : list
        The best parameters identified by SPSA.

    """
    if len(x0)==0:
        x0 = bm.sample()
    
    counter = 0
    for variables in spsa.iterator.minimize(spsa.with_input_noise(bm.cost, shape=bm.n_parameters(), noise = 0.01), x0, adam=False):
        if counter % printFrequency == 0:
            print('-----------')
            print('Current cost:')
            print(variables['y'])
            print('Found at point:')
            print(variables['x'])
            print('Best cost:')
            print(variables['y_best'])
            print('Found at point:')
            print(variables['x_best'])
        counter += 1
        if counter > maxiter:
            break
    
    bm.evaluate(variables['x_best'])
    return variables['x_best']

if __name__ == '__main__':
    bm = staircase.HH_Benchmarker()
    x0 = bm.defaultParams
    run(bm, x0)


import pints
import ionbench
import numpy as np
from ionbench.optimisers.pints_optimisers import classes_pints

def run(bm, x0 = [], maxIter=1000):
    """
    Runs CMA-ES from Pints using a benchmarker. 

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is [], in which case a randomly sampled parameter vector is retrieved from bm.sample().
    maxIter : int, optional
        Number of iterations of CMA-ES to run. The default is 1000.

    Returns
    -------
    xbest : list
        The best parameters identified by CMA-ES.

    """
    if len(x0)==0:
        x0 = bm.sample()
    model = classes_pints.Model(bm)
    problem = pints.SingleOutputProblem(model, np.arange(model.bm.tmax), model.bm.data)
    error = pints.RootMeanSquaredError(problem)
    if bm._bounded:
        boundaries = classes_pints.AdvancedBoundaries(bm)
    
    if bm._bounded:
        counter = 1
        while not boundaries.check(x0):
            x0 = bm.sample()
            counter += 1
        if counter > 10:
            print("Struggled to find parameters in bounds")
            print("Required "+str(counter)+" iterations")
    # Create an optimisation controller
    if bm._bounded:
        opt = pints.OptimisationController(error, x0, method=pints.CMAES, boundaries = boundaries)
    else:
        opt = pints.OptimisationController(error, x0, method=pints.CMAES)
    opt.set_max_iterations(maxIter)
    # Run the optimisation
    x, f = opt.run()
    
    model.bm.evaluate(x)
    return x

if __name__ == '__main__':
    bm = ionbench.problems.staircase.HH_Benchmarker()
    bm.log_transform([True, False]*4+[False])
    bm.add_bounds([[1e-7]*8+[-np.inf], [1e3,0.4]*4+[np.inf]])
    run(bm)

def get_approach():
    """
    The approach in Clerx et al 2019. Uses log transforms, and bounds (technically they are bounds on rates here. We just use sampler bounds for now).

    Returns
    -------
    app : approach
        The approach used in Clerx et al 2019.

    """
    app = ionbench.approach.Clerx2019()
    return app
